#include <windows.h>
#include <process.h>
#include <stdio.h>
#include <locale.h>
#include <tchar.h>
#include <vector>

#ifndef ARRAYSIZE
    #define ARRAYSIZE(ary) (sizeof(ary) / sizeof(ary[0]))
#endif

BOOL g_bCtrlC = FALSE;

unsigned __stdcall WatchThreadProc(void *arg)
{
    LPCSTR pszFileName = (LPCSTR)arg;
    char buffer[0x1E00];
    HANDLE hDir;
	OVERLAPPED overlapped;
    BOOL fRecursive = TRUE;
	DWORD ret_size;
    HANDLE hEvent;
    FILE_NOTIFY_INFORMATION *pInfo;
    DWORD dwFilter = FILE_NOTIFY_CHANGE_FILE_NAME   |
                     FILE_NOTIFY_CHANGE_DIR_NAME    |
                     FILE_NOTIFY_CHANGE_ATTRIBUTES  |
                     FILE_NOTIFY_CHANGE_SIZE        |
                     FILE_NOTIFY_CHANGE_LAST_WRITE  |
                     FILE_NOTIFY_CHANGE_CREATION;
    const char *action;
    WCHAR szName[MAX_PATH];
    CHAR szNameA[MAX_PATH];
    DWORD cbName;

    // creaate manual-reset event
	hEvent = CreateEventA(NULL, TRUE, FALSE, NULL);
    if (hEvent == NULL)
    {
        printf("CreateEventA failed: 0x%08lX\n", GetLastError());
        return -2;
    }

    // open the directory to watch
    hDir = CreateFileA(pszFileName,
                       FILE_LIST_DIRECTORY,
                       FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
                       NULL);
    if (hDir == INVALID_HANDLE_VALUE)
    {
        printf("CreateFileA failed: 0x%08lX\n", GetLastError());
        return -1;
    }

    // wait loop
    for (;;)
    {
        // manual rest
        ResetEvent(hEvent);

        // initialize OVERLAPPED
        ZeroMemory(&overlapped, sizeof(overlapped));
        overlapped.hEvent = hEvent;

        // start dir watch
        if (!ReadDirectoryChangesW(hDir, buffer, sizeof(buffer),
                                   fRecursive, dwFilter, NULL, &overlapped, NULL))
        {
            printf("ReadDirectoryChangesW failed: 0x%08lX\n", GetLastError());
            break;
        }

        // wait for change
        while (!g_bCtrlC)
        {
            DWORD wait = WaitForSingleObject(hEvent, 500);
            if (wait != WAIT_TIMEOUT)
                break;
            printf(".");
        }

        // quit now?
        if (g_bCtrlC)
        {
            // cancel async I/O
            CancelIo(hDir);

            // wait for cancel
            WaitForSingleObject(hEvent, INFINITE);
            break;
        }

        // get result
        ret_size = 0;
        if (!GetOverlappedResult(hDir, &overlapped, &ret_size, FALSE))
        {
            printf("GetOverlappedResult failed: 0x%08lX\n", GetLastError());
            break;
        }

        printf("\nret_size: %lu\n", ret_size);

        if (ret_size == 0)
        {
            printf("buffer overflow\n");
            continue;
        }

        // read changes
        pInfo = (FILE_NOTIFY_INFORMATION *)buffer;
        for (;;)
        {
            // get action name
            action = "(unknown)";
            switch (pInfo->Action)
            {
            case FILE_ACTION_ADDED: action = "added"; break;
            case FILE_ACTION_REMOVED: action = "removed"; break;
            case FILE_ACTION_MODIFIED: action = "modified"; break;
            case FILE_ACTION_RENAMED_OLD_NAME: action = "renamed old"; break;
            case FILE_ACTION_RENAMED_NEW_NAME: action = "renamed new"; break;
            }

            // get name and output
            // NOTE: FILE_NOTIFY_INFORMATION.FileName is not null-terminated.
            cbName = pInfo->FileNameLength;
            if (sizeof(szName) - sizeof(UNICODE_NULL) >= cbName)
            {
                ZeroMemory(szName, sizeof(szName));
                CopyMemory(szName, pInfo->FileName, cbName);

                WideCharToMultiByte(CP_ACP, 0, szName, -1, szNameA, ARRAYSIZE(szNameA), NULL, NULL);
                printf("%s: %s\n", action, szNameA);
            }

            if (pInfo->NextEntryOffset == 0) // no next
                break;

            // get next
            pInfo = (FILE_NOTIFY_INFORMATION *)((LPBYTE)pInfo + pInfo->NextEntryOffset);
        }
    }

    CloseHandle(hEvent);
    CloseHandle(hDir);

    return 0;
}

BOOL CALLBACK HandlerRoutine(DWORD dwCtrlType)
{
    switch (dwCtrlType)
    {
    case CTRL_C_EVENT:
        g_bCtrlC = TRUE;
        return TRUE;
    }
    return FALSE;
}

int main(int argc, char **argv)
{
    char dir[MAX_PATH] = ".";
    if (argc >= 2)
        lstrcpynA(dir, argv[1], ARRAYSIZE(dir));

    setlocale(LC_ALL, "");
    SetConsoleCtrlHandler(HandlerRoutine, TRUE);
    printf("Press Ctrl+C to quit\n\n");

    unsigned tid;
    HANDLE hThread = (HANDLE)_beginthreadex(NULL, 0x2000, WatchThreadProc, dir, 0, &tid);
    WaitForSingleObject(hThread, INFINITE);
    CloseHandle(hThread);

    printf("quit\n\n");

    return 0;
}
